"""Utilities used in tests.
"""


class RaisingConverter(object):
    """Used to replace the ftw.tika.converter.TikaConverter with a thing that
    conforms to the same interface, but just raises a defined exception once
    the convert() method is called.
    """

    def __init__(self, exception):
        self.exception = exception

    def convert(self, doc, *args, **kwargs):
        raise self.exception



"""MIME types supported by ftw.tika.

If you update any type definitions in here, please make sure you also update
the README accordingly.
"""

PDF_TYPES = [
    # PDF documents
    'application/pdf',
    'application/x-pdf',
    'image/pdf',
    'application/x-gzpdf',
    'application/x-bzpdf',
    'application/postscript',
    'application/x-gzpostscript',
    ]


TYPES = [
# Microsoft Office formats (Office Open XML)
'application/vnd.openxmlformats-officedocument.presentationml.presentation',
'application/vnd.openxmlformats-officedocument.presentationml.slide',
'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
'application/vnd.openxmlformats-officedocument.presentationml.template',
'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
'application/vnd.openxmlformats-officedocument.wordprocessingml.template',

# OpenOffice ODF formats
'application/vnd.oasis.opendocument.chart',
'application/vnd.oasis.opendocument.chart-template',
'application/vnd.oasis.opendocument.database',
'application/vnd.oasis.opendocument.formula',
'application/vnd.oasis.opendocument.formula-template',
'application/vnd.oasis.opendocument.graphics',
'application/vnd.oasis.opendocument.graphics-template',
'application/vnd.oasis.opendocument.image',
'application/vnd.oasis.opendocument.image-template',
'application/vnd.oasis.opendocument.presentation',
'application/vnd.oasis.opendocument.presentation-template',
'application/vnd.oasis.opendocument.spreadsheet',
'application/vnd.oasis.opendocument.spreadsheet-template',
'application/vnd.oasis.opendocument.text',
'application/vnd.oasis.opendocument.text-master',
'application/vnd.oasis.opendocument.text-template',
'application/vnd.oasis.opendocument.text-web',

# OpenOffice 1.x formats
'application/vnd.sun.xml.calc',
'application/vnd.sun.xml.calc.template',
'application/vnd.sun.xml.draw',
'application/vnd.sun.xml.draw.template',
'application/vnd.sun.xml.impress',
'application/vnd.sun.xml.impress.template',
'application/vnd.sun.xml.math',
'application/vnd.sun.xml.writer',
'application/vnd.sun.xml.writer.global',
'application/vnd.sun.xml.writer.template',

# Legacy Microsoft Office (97) formats
'application/vnd.ms-excel',
'application/msexcel',
'application/x-msexcel',
'application/vnd.ms-powerpoint',
'application/powerpoint',
'application/mspowerpoint',
'application/x-mspowerpoint',
'application/vnd.ms-works',
'application/msword',
'application/msword-template',
'application/rtf',
'application/vnd.visio',
'application/vnd.ms-tnef',
'application/ms-tnef',

# Adobe formats
'application/x-indesign',
'application/illustrator',
'image/x-photoshop',

# Other office formats
'application/vnd.wordperfect',
'application/x-wordperfect',
'application/wordperfect',

# E-Mail messages
'message/rfc822',
] + PDF_TYPES


MS_OFFICE_TYPES = [
    mt for mt in TYPES
    if mt.startswith('application/vnd.openxmlformats-officedocument.')]



from ftw.tika.interfaces import IZCMLTikaConfig
from ftw.tika.testing import META_ZCML
from unittest2 import TestCase
from zope.component import getUtility
from zope.component import queryUtility
from zope.configuration.exceptions import ConfigurationError


class TestTikaConfigDirective(TestCase):

    layer = META_ZCML

    def test_no_config_utility_registered_when_not_configured(self):
        self.assertFalse(queryUtility(IZCMLTikaConfig),
                         'Expected IZCMLTikaConfig to not be registered yet.')

    def test_zcml_directive_registers_utility(self):
        self.load_zcml('<tika:config />')
        self.assertTrue(queryUtility(IZCMLTikaConfig),
                        'IZCMLTikaConfig utility should be registered after'
                        ' configuring it in ZCML.')

    def test_config_stores_path(self):
        self.load_zcml('<tika:config path="/path/to/tika-app.jar" />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals('/path/to/tika-app.jar', config.path)

    def test_config_stores_port(self):
        self.load_zcml('<tika:config port="9998" />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals(9998, config.port)

    def test_port_needs_to_be_integer(self):
        with self.assertRaises(ConfigurationError) as cm:
            self.load_zcml('<tika:config port="foo" />')
        self.assertIn('ValueError: invalid literal for int',
                      str(cm.exception))

    def test_config_stores_host(self):
        self.load_zcml('<tika:config host="other.host" />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals('other.host', config.host)

    def test_host_defaults_to_localhost(self):
        self.load_zcml('<tika:config />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals('localhost', config.host)

    def test_config_stores_timeout(self):
        self.load_zcml('<tika:config timeout="50" />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals(50, config.timeout)

    def test_config_defaults_timeout(self):
        self.load_zcml('<tika:config />')
        config = getUtility(IZCMLTikaConfig)
        self.assertEquals(10, config.timeout)

    def test_timeout_needs_to_be_integer(self):
        with self.assertRaises(ConfigurationError) as cm:
            self.load_zcml('<tika:config timeout="foo" />')
        self.assertIn('ValueError: invalid literal for int',
                      str(cm.exception))

    def load_zcml(self, *lines):
        self.layer.load_zcml_string('\n'.join((
                    '<configure xmlns:tika="http://namespaces.plone.org/tika">',
                    ) + lines +  (
                    '</configure>',
                    )))



"""Utility functions for detecting password protected documents upon
conversion failure.
"""

from ftw.tika import mimetypes


PROTECTED_PDF_MSGS = (
    # Tika 1.5, 1.6
    'Error: The supplied password does not match either the'
    ' owner or user password in the document.',
    # Tika 1.7
    'Document is encrypted',
)

PROTECTED_MSOFFICE_MSGS = (
    # Tika 1.5, 1.6, 1.7
    'org.apache.tika.exception.EncryptedDocumentException',
)


def is_protected_doc(exc, mimetype):
    """Figure out whether conversion failed because we're dealing with
    a password protected document, based on the document's MIME type and
    the Java exception.
    """
    stack_trace = str(exc.stack_trace)
    if not stack_trace:
        # No Java stack trace to aid in the detection of the exact error
        return False

    if mimetype in mimetypes.PDF_TYPES:
        return any(msg in stack_trace for msg in PROTECTED_PDF_MSGS)

    if mimetype in mimetypes.MS_OFFICE_TYPES:
        return any(msg in stack_trace for msg in PROTECTED_MSOFFICE_MSGS)

    return False



from ftw.tika.exceptions import ProcessError
from ftw.tika.exceptions import TikaConversionError
from ftw.tika.exceptions import TikaJarNotConfigured
from ftw.tika.exceptions import TikaJarNotFound
from ftw.tika.interfaces import IZCMLTikaConfig
from ftw.tika.utils import clean_extracted_plaintext
from ftw.tika.utils import run_process
from plone.memoize import instance
from requests.exceptions import RequestException
from StringIO import StringIO
from threading import local
from zope.component import queryUtility
import logging
import os
import requests
import socket
import tempfile


thread_locals = local()


def get_requests_session():
    attrname = 'tika_requests_session'
    if not hasattr(thread_locals, attrname):
        setattr(thread_locals, attrname, requests.session())
    return getattr(thread_locals, attrname)


def copy_stream(input_, output):
    """Reads from the ``input_`` stream or string and writes into
    the ``output``. It does this in a buffered fashion for making
    sure that we never hold the whole file in our memory.

    This is used for copying the file data from Plone into either
    a tempfile when tika is used locally or into the socket which
    connects to the tika server.
    """
    if isinstance(input_, basestring):
        output.write(input_)
        return

    while True:
        data = input_.read(1024)
        if not data:
            break
        output.write(data)


class TikaConverter(object):
    """Converts documents to text/plain using Apache Tika.
    """

    def __init__(self, path=None):
        self._path = path
        self.log = logging.getLogger('ftw.tika')

    @property
    def java_path(self):
        return 'java'

    @property
    @instance.memoize
    def config(self):
        return queryUtility(IZCMLTikaConfig)

    @property
    def jar_path(self):
        # Try path supplied to constructor first
        if self._path not in (None, ''):
            path = self._path
        else:
            # Otherwise try to get path from ZCML
            if self.config is None or self.config.path is None:
                # Bail if neither are provided
                msg = "No path to Tika JAR file specified."
                raise TikaJarNotConfigured(msg)
            else:
                path = self.config.path

        # Check if the specified path is actually a file
        path = os.path.abspath(path)
        if not os.path.isfile(path):
            msg = "No such file: %s" % path
            raise TikaJarNotFound(msg)
        return path

    @property
    def server_configured(self):
        if self.config is None:
            return False
        return self.config.port is not None

    def convert(self, document, filename=''):
        """Converts `document` to 'text/plain' using Apache Tika.

        `document` can either be an string or a file-like stream object.
        """

        if self.server_configured:
            try:
                text = self.convert_server(document, filename)

            except requests.exceptions.ConnectionError, exc:
                # Could not connect to the server at all.
                # Server is probably not running
                self.log.error(
                    'Could not connect to tika server: %s' % str(exc))
                text = self.convert_local(document, filename)

            except (socket.error, RequestException), exc:
                # An error occured, maybe a connection timeout.
                self.log.error(
                    'Full text extraction using tika failed: %s' % str(exc))
                text = u''
        else:
            text = self.convert_local(document, filename)

        return text

    def convert_server(self, document, filename=''):
        base_url = "http://{0}:{1}".format(self.config.host, self.config.port)
        tika_endpoint = '/'.join((base_url, 'tika'))
        self.log.info(
            'Converting document with tika JAXRS server: %s' % filename)

        if isinstance(document, basestring):
            document = StringIO(document)

        headers = {'Accept': 'text/plain'}
        timeout = self.config.timeout
        response = get_requests_session().put(tika_endpoint,
                                              data=document,
                                              headers=headers,
                                              timeout=timeout)

        status, body = response.status_code, response.content

        if not status == 200:
            msg = ("Conversion with Tika JAXRS server failed "
                   "with status %s. " % status)
            raise TikaConversionError(
                msg, status_code=status, stack_trace=body.strip())

        text = clean_extracted_plaintext(body, filename)
        return text

    def convert_local(self, document, filename=''):
        self.log.info('Converting document with LOCAL tika: %s' % filename)
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        copy_stream(document, temp_file)
        temp_file.close()

        try:
            cmd = ' '.join([self.java_path, '-jar', self.jar_path,
                            '-t', temp_file.name])
            try:
                stdout, stderr = run_process(cmd)
            except ProcessError, e:
                msg = "Conversion with local Tika failed."
                stack_trace = e.message
                raise TikaConversionError(msg, stack_trace=stack_trace)

            text = clean_extracted_plaintext(stdout, filename)
            return text

        finally:
            os.unlink(temp_file.name)



from ftw.tika.converter import TikaConverter
from ftw.tika.testing import FTW_TIKA_INTEGRATION_TESTING
from unittest2 import TestCase
import os.path


ASSETS = os.path.join(os.path.dirname(__file__), 'assets')


class TestLocalConversion(TestCase):

    layer = FTW_TIKA_INTEGRATION_TESTING

    def test_convert_docx(self):
        converter = TikaConverter()
        with open(os.path.join(ASSETS, 'lorem.docx')) as file_:
            result = converter.convert_local(file_)
        self.assertEquals('Lorem Ipsum', result.strip())



from ftw.tika.converter import TikaConverter
from ftw.tika.testing import TIKA_SERVER_INTEGRATION_TESTING
from ftw.tika.utils import strip_word_bookmarks
from unittest2 import TestCase
import os.path


ASSETS = os.path.join(os.path.dirname(__file__), 'assets')


class TestServerConversion(TestCase):

    layer = TIKA_SERVER_INTEGRATION_TESTING

    def test_convert_docx(self):
        converter = TikaConverter()
        filename = 'lorem.docx'
        with open(os.path.join(ASSETS, filename)) as file_:
            result = converter.convert_server(file_)
        self.assertEquals(
            'Lorem Ipsum',
            strip_word_bookmarks(result.strip(), filename=filename))



"""Exceptions raised by ftw.tika.
"""


class ProcessError(Exception):
    """Raised if a spawned process terminated with exit code != 0.
    """


class TikaConversionError(Exception):
    """Raised if an error happened during conversion with Tika.

    May contain additional information about the error on the Tika side:
    - status_code: The HTTP status code returned by the Tika JAXRS server
    - stack_trace: The Java stack trace from the tika-app.jar, or, if
      available, from the Tika JAXRS Server
    """
    def __init__(self, message, status_code=None, stack_trace=None):
        self.message = message
        self.status_code = status_code
        self.stack_trace = stack_trace

    def __str__(self):
        return self.message


class TikaJarNotConfigured(Exception):
    """Raised if no path to Tika JAR file was specified.
    """


class TikaJarNotFound(Exception):
    """Raised if a path to Tika JAR file was specified, but is invalid.
    """



from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite
import os.path


ASSETS = os.path.join(os.path.dirname(__file__), 'assets')
OFFICE_MIME = 'application/vnd.openxmlformats-officedocument.'
MIMETYPES_BY_EXTENSION = {
    '.docx': '%swordprocessingml.document' % OFFICE_MIME,
    '.doc': 'application/msword',
    '.xlsx': '%sspreadsheetml.sheet' % OFFICE_MIME,
    '.xls': 'application/msexcel',
    '.pptx': '%spresentationml.presentation' % OFFICE_MIME,
    '.ppt': 'application/mspowerpoint',
    '.rtf': 'application/rtf',
    '.odt': 'application/vnd.oasis.opendocument.text',
    '.sxw': 'application/vnd.sun.xml.writer',
    '.eml': 'message/rfc822',
    '.pdf': 'application/pdf'}


def convert_asset(filename):
    path = os.path.join(ASSETS, filename)
    transforms = getToolByName(getSite(), 'portal_transforms')
    _, extension = os.path.splitext(filename)
    mimetype = MIMETYPES_BY_EXTENSION[extension]
    with open(path, 'r') as asset:
        stream = transforms.convertTo('text/plain', asset, mimetype=mimetype,
                                      filename=filename)
        return stream.getData().strip()



from ftw.tika.exceptions import ProcessError
from ftw.tika.utils import run_process
from ftw.tika.utils import strip_thumbnail_names
from ftw.tika.utils import strip_word_bookmarks
from unittest2 import TestCase


class TestUtils(TestCase):

    def test_run_process_returns_stdout_and_stderr(self):
        cmd = "echo stdout; echo stderr 1>&2"
        stdout, stderr = run_process(cmd)
        self.assertEquals(stdout, 'stdout\n')
        self.assertEquals(stderr, 'stderr\n')

    def test_run_process_raises_process_error_on_nonzero_exit_code(self):
        cmd = "false"
        with self.assertRaises(ProcessError):
            stdout, stderr = run_process(cmd)

    def test_strip_word_bookmarks_removes_bookmarks_from_docx(self):
        text = '[bookmark: Foo][bookmark: _GoBack]Lorem Ipsum[bookmark: Foo]'
        self.assertEquals(
            'Lorem Ipsum',
            strip_word_bookmarks(text, filename='foo.docx'))

    def test_strip_word_bookmarks_doesnt_touch_files_other_than_docx(self):
        text = '[bookmark: Foo][bookmark: _GoBack]Lorem Ipsum[bookmark: Foo]'
        self.assertEquals(text, strip_word_bookmarks(text))

    def test_strip_thumbnail_names_removes_thumbnail_filenames(self):
        text = 'Lorem Ipsum\nthumbnail_0.jpeg'
        self.assertEquals(
            'Lorem Ipsum',
            strip_thumbnail_names(text).strip())



from ftw.testing import ComponentRegistryLayer
from ftw.tika.interfaces import IZCMLTikaConfig
from plone.app.testing import applyProfile
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles, TEST_USER_ID, TEST_USER_NAME, login
from plone.testing import Layer
from subprocess import Popen
from threading import Thread
from zope.component import getUtility
from zope.configuration import xmlconfig
import os
import time


def tika_version():
    """Determine the Tika version we're testing against.
    """
    version = os.environ.get('TESTING_TIKA_VERSION', '0.0')
    return tuple(map(int, version.split('.')))


class MetaZCMLLayer(ComponentRegistryLayer):

    def setUp(self):
        super(MetaZCMLLayer, self).setUp()
        import ftw.tika
        self.load_zcml_file('meta.zcml', ftw.tika)


META_ZCML = MetaZCMLLayer()


class FtwTikaLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        import ftw.tika
        xmlconfig.file('configure.zcml', ftw.tika,
                       context=configurationContext)

        # os.getcwd() -> .../parts/test
        parts_path = os.path.join(os.getcwd(), '..')

        app_path = os.path.join(
            parts_path, 'tika-app-download', 'tika-app.jar')
        self['tika_config'] = {'path': app_path,
                               'port': os.environ.get('PORT1', '55007')}

        xmlconfig.string(
            '<configure xmlns:tika="http://namespaces.plone.org/tika">' +
            '<tika:config path="%(path)s" />' % (
                self['tika_config']) +
            '</configure>',
            context=configurationContext)

        self['server_path'] = os.path.join(
            parts_path, 'tika-server-download', 'tika-server.jar')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.tika:default')

        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)


FTW_TIKA_FIXTURE = FtwTikaLayer()
FTW_TIKA_INTEGRATION_TESTING = IntegrationTesting(
    bases=(FTW_TIKA_FIXTURE,),
    name="FtwTika:Integration")


class TikaServerLayer(Layer):

    defaultBases = (FTW_TIKA_FIXTURE, )

    def setUp(self):
        self.start_server()

    def tearDown(self):
        self.stop_server()

    def testSetUp(self):
        # We explicitly remove the path for disabling fallback to
        # standalone method, so that we can make sure that the server
        # is actually used in this layer.
        tika_config = getUtility(IZCMLTikaConfig)
        tika_config.path = None
        tika_config.port = int(self['tika_config']['port'])

    def testTearDown(self):
        tika_config = getUtility(IZCMLTikaConfig)
        tika_config.path = self['tika_config']['path']
        tika_config.port = None

    def start_server(self):
        srv_config = {'server_path': self['server_path']}
        srv_config.update(self['tika_config'])

        additional_opts = []

        if tika_version() >= (1, 8):
            # Tika 1.8 allows to include the Java stack traces in the
            # response body for failed conversion requests.
            additional_opts.append('-includeStack')

        srv_config['additional_opts'] = ' '.join(additional_opts)

        command = ('java -jar %(server_path)s --port %(port)s '
                   '%(additional_opts)s' % (srv_config))

        self.process = Popen(command, shell=True)
        Thread(target=self.process.communicate).start()
        time.sleep(4.0)  # give tika some time to boot

    def stop_server(self):
        if getattr(self, 'process', None) is not None:
            self.process.terminate()

TIKA_SERVER_FIXTURE = TikaServerLayer()
TIKA_SERVER_INTEGRATION_TESTING = IntegrationTesting(
    bases=(TIKA_SERVER_FIXTURE,),
    name='FtwTika:server:Integration')



from ftw.tika.interfaces import IZCMLTikaConfig
from zope import schema
from zope.component.zcml import utility
from zope.interface import implements
from zope.interface import Interface


class ZCMLTikaConfig(object):
    implements(IZCMLTikaConfig)

    def __init__(self, path=None, port=None, host='localhost', timeout=10):
        self.path = path
        self.port = port
        self.host = host
        self.timeout = timeout


class ITikaConfigDirective(Interface):
    """Directive which registers a Tika config.
    """

    path = schema.ASCIILine(
        title=u"tika-app JAR Path",
        description=u"Path to the tika-app JAR file.",
        required=False,
    )

    port = schema.Int(
        title=u'Server port',
        description=u'The tika server port.',
        required=False)

    host = schema.ASCIILine(
        title=u"Server host",
        description=u'The tika server host.',
        required=False,
        )

    timeout = schema.Int(
        title=u'Connection Timeout',
        description=u'The timeout for the Connection to the tika server.',
        required=False,
        default=10)


def tikaConfigDirective(_context, **arguments):
    """The <tika:config /> directive.
    Usage:

    <configure xmlns:tika="http://namespaces.plone.org/tika">
        <tika:config path="/path/to/tika-app.jar"
                     host="tika.host"
                     port="9998"
                     timeout="10" />
    </configure>
    """

    utility(_context,
            provides=IZCMLTikaConfig,
            component=ZCMLTikaConfig(**arguments))



from ftw.tika.mimetypes import TYPES
from ftw.tika.setuphandlers import RegistrationUtility
from ftw.tika.testing import FTW_TIKA_INTEGRATION_TESTING
from ftw.tika.transforms.tika_to_plain_text import TIKA_TRANSFORM_NAME
from Products.CMFCore.utils import getToolByName
from Products.PortalTransforms.utils import TransformException
from unittest2 import TestCase
import logging


logger = logging.getLogger('ftw.tika.tests')


class TestInstallation(TestCase):

    layer = FTW_TIKA_INTEGRATION_TESTING

    def test_default_profile_registers_transform(self):
        portal = self.layer['portal']
        portal_transforms = getToolByName(portal, 'portal_transforms')
        self.assertIn('tika_to_plain_text', portal_transforms.keys())

    def test_transforms_input_types_are_complete_after_installation(self):
        portal = self.layer['portal']
        portal_transforms = getToolByName(portal, 'portal_transforms')
        util = RegistrationUtility(portal, None)

        transform = portal_transforms[TIKA_TRANSFORM_NAME]
        filtered_types = list(util.filter_types_to_registered_ones_only(TYPES))
        self.assertEquals(set(filtered_types), set(transform.inputs))

    def test_running_default_profile_twice_shouldnt_cause_errors(self):
        portal = self.layer['portal']
        setup_tool = getToolByName(portal, 'portal_setup')

        setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:default',
                                                purge_old=False)

        # Running the uninstall profile twice shouldn't cause any errors
        setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:default',
                                                purge_old=False)

    def test_attempting_to_install_with_existing_transform_policy_raises(self):
        portal = self.layer['portal']
        setup_tool = getToolByName(portal, 'portal_setup')
        #transform_tool = getUtility(IPortalTransformsTool)

        util = RegistrationUtility(portal, logger)
        util.unregister_transform_policy("text/plain")
        util.register_transform_policy("text/plain", 'some_transform')

        with self.assertRaises(TransformException):
            setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:default',
                                                    purge_old=False)

    def test_uninstall_profile_removes_transform(self):
        portal = self.layer['portal']
        setup_tool = getToolByName(portal, 'portal_setup')
        portal_transforms = getToolByName(portal, 'portal_transforms')
        setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:uninstall',
                                                purge_old=False)
        self.assertNotIn('tika_to_plain_text', portal_transforms.keys())

    def test_running_uninstall_profile_twice_shouldnt_cause_errors(self):
        portal = self.layer['portal']
        setup_tool = getToolByName(portal, 'portal_setup')

        setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:uninstall',
                                                purge_old=False)

        # Running the uninstall profile twice shouldn't cause any errors
        setup_tool.runAllImportStepsFromProfile('profile-ftw.tika:uninstall',
                                                purge_old=False)



from ftw.tika.testing import FTW_TIKA_INTEGRATION_TESTING
from ftw.tika.testing import TIKA_SERVER_INTEGRATION_TESTING
from ftw.tika.testing import tika_version
from ftw.tika.tests.helpers import convert_asset
from testfixtures import log_capture
from unittest2 import TestCase


PROTECTED_MSG = (
    'ftw.tika', 'INFO', 'Could not convert password protected document.')


class TestConversion(TestCase):

    layer = FTW_TIKA_INTEGRATION_TESTING

    def test_docx_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.docx'))

    def test_doc_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.doc'))

    def test_xlsx_conversion(self):
        self.assertEquals('Sheet1\n\tLorem Ipsum',
                          convert_asset('lorem.xlsx'))

    def test_xls_conversion(self):
        self.assertEquals('Sheet1\n\tLorem Ipsum',
                          convert_asset('lorem.xls'))

    def test_pptx_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.pptx'))

    def test_ppt_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.ppt'))

    def test_pdf_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.pdf'))

    def test_rtf_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.rtf'))

    def test_odt_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.odt'))

    def test_sxw_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.sxw'))

    def test_eml_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.eml'))

    @log_capture('ftw.tika')
    def test_protected_pdf_conversion(self, log):
        self.assertEquals('', convert_asset('protected.pdf'))
        self.assertIn(PROTECTED_MSG, tuple(log.actual()))

    @log_capture('ftw.tika')
    def test_protected_docx_conversion(self, log):
        self.assertEquals('', convert_asset('protected.docx'))
        self.assertIn(PROTECTED_MSG, tuple(log.actual()))


class TestServerConversion(TestCase):

    layer = TIKA_SERVER_INTEGRATION_TESTING

    def test_docx_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.docx'))

    def test_doc_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.doc'))

    def test_xlsx_conversion(self):
        self.assertEquals('Sheet1\n\tLorem Ipsum',
                          convert_asset('lorem.xlsx'))

    def test_xls_conversion(self):
        self.assertEquals('Sheet1\n\tLorem Ipsum',
                          convert_asset('lorem.xls'))

    def test_pptx_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.pptx'))

    def test_ppt_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.ppt'))

    def test_pdf_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.pdf'))

    def test_rtf_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.rtf'))

    def test_odt_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.odt'))

    def test_sxw_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.sxw'))

    def test_eml_conversion(self):
        self.assertEquals('Lorem Ipsum', convert_asset('lorem.eml'))

    @log_capture('ftw.tika')
    def test_protected_pdf_conversion(self, log):
        self.assertEquals('', convert_asset('protected.pdf'))
        if tika_version() >= (1, 8):
            self.assertIn(PROTECTED_MSG, tuple(log.actual()))
        else:
            # Assertion on status 422 doesn't work reliably - sometimes Tika
            # fails with a 500, sometimes with 422 (for the same test)
            self.assertIn(
                'Conversion with Tika JAXRS server failed with status',
                str(tuple(log.actual())))

    @log_capture('ftw.tika')
    def test_protected_docx_conversion(self, log):
        self.assertEquals('', convert_asset('protected.docx'))
        if tika_version() >= (1, 8):
            self.assertIn(PROTECTED_MSG, tuple(log.actual()))
        else:
            # Assertion on status 422 doesn't work reliably - sometimes Tika
            # fails with a 500, sometimes with 422 (for the same test)
            self.assertIn(
                'Conversion with Tika JAXRS server failed with status',
                str(tuple(log.actual())))



"""Utility functions for ftw.tika.
"""

from ftw.tika.exceptions import ProcessError
import re
import subprocess


def run_process(cmd):
        process = subprocess.Popen(cmd,
                                   shell=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)

        # wait for the process to terminate
        stdout, stderr = process.communicate()
        errcode = process.returncode

        if errcode != 0:
            raise ProcessError(stderr)

        return (stdout, stderr)


def clean_extracted_plaintext(text, filename=''):
    """Clean up raw plain text returned by either Tika Server or a local tika
    app by removing common noise that we don't want in the extracted text,
    like word bookmark names or filenames of thumbnails.
    """
    text = strip_word_bookmarks(text, filename)
    text = strip_thumbnail_names(text, filename)
    return text


def strip_word_bookmarks(text, filename=''):
    """Newer versions of tika also include Word bookmarks in the returned
    plain text. They are inserted in the form
    '[bookmark: Foo][bookmark: _GoBack]Lorem Ipsum'

    For our purpose of building the searchable text, we want to strip those.
    """
    if not filename.lower().endswith('.docx'):
        return text

    pattern = re.compile(r'\[bookmark:.*?\]')
    text = re.sub(pattern, '', text)
    return text


def strip_thumbnail_names(text, filename=''):
    """Since Tika 1.6 filenames of thumbnails in office documents also get
    returned as part of the plain text. Therefore we strip the common
    default filenames for thumbnails here.
    """
    pattern = re.compile(r'thumbnail_[0-9*]\.jpeg')
    text = re.sub(pattern, '', text)
    return text



from ftw.tika.converter import TikaConverter
from ftw.tika.exceptions import TikaConversionError
from ftw.tika.protected_docs import is_protected_doc
from Products.PortalTransforms.interfaces import ITransform
from ZODB.POSException import ConflictError
from zope.interface import implements
import logging
import os


TIKA_TRANSFORM_NAME = 'tika_to_plain_text'
INITIAL_TYPES = ['application/pdf']

logger = logging.getLogger('ftw.tika')


class Tika2TextTransform(object):
    """A portal transform that converts various office document formats to
    'text/plain' using Apache Tika.
    """

    implements(ITransform)

    __name__ = TIKA_TRANSFORM_NAME
    output = "text/plain"

    def __init__(self, name=None, inputs=INITIAL_TYPES):
        self.config = {'inputs': inputs}
        self.config_metadata = {
            'inputs': ('list', 'Inputs',
                       'Input(s) MIME type. Change with care.'),
            }
        if name:
            self.__name__ = name

    def name(self, *args):
        return self.__name__

    def __getattr__(self, attr):
        if attr in self.config:
            return self.config[attr]
        raise AttributeError(attr)

    def convert(self, orig, data, filename='', mimetype=None, **kwargs):
        converter = TikaConverter()
        try:
            plain_text = converter.convert(orig, filename=filename)

        except (ConflictError, KeyboardInterrupt):
            raise

        except TikaConversionError, exc:
            self._log_conversion_error(exc, mimetype=mimetype)
            plain_text = ''

        except Exception, exc:
            logger.warn(exc)
            plain_text = ''

        data.setData(plain_text)
        return data

    def _log_conversion_error(self, exc, mimetype):
        if is_protected_doc(exc, mimetype):
            logger.info('Could not convert password protected document.')
        else:
            logger.warn(exc)
            if os.environ.get('FTW_TIKA_VERBOSE_LOGGING', False):
                msg = "Tika output:\n%s" % str(exc.stack_trace)
                logger.warn(msg)


def register():
    return Tika2TextTransform()



from ftw.tika.mimetypes import TYPES
from ftw.tika.transforms.tika_to_plain_text import Tika2TextTransform
from ftw.tika.transforms.tika_to_plain_text import TIKA_TRANSFORM_NAME
from Products.CMFCore.utils import getToolByName
from Products.PortalTransforms.interfaces import IPortalTransformsTool
from Products.PortalTransforms.utils import TransformException
from zope.component import getUtility


# The profile id of your package:
PROFILE_ID = 'profile-ftw.tika:default'


class RegistrationUtility(object):
    """Handles registering and unregistering portal transforms and their
    policies.
    """

    def __init__(self, context, logger):
        self.context = context
        self.logger = logger

    def register_transform(self, transform):
        transform_tool = getUtility(IPortalTransformsTool)
        transform = transform()

        self.logger.info("Registering portal transform "
                         "'%s' " % transform.__name__)
        transform_tool.registerTransform(transform)

    def unregister_transform(self, transform_name):
        transform_tool = getUtility(IPortalTransformsTool)
        if hasattr(transform_tool, transform_name):
            self.logger.info("Unregistering portal transform "
                             "'%s' " % transform_name)
            transform_tool.unregisterTransform(transform_name)

    def register_transform_policy(self, output_mimetype, required_transform):
        transform_tool = getUtility(IPortalTransformsTool)
        transform_tool.manage_addPolicy(output_mimetype,
                                        [required_transform])

    def unregister_transform_policy(self, output_mimetype):
        transform_tool = getUtility(IPortalTransformsTool)
        policies = [mimetype for (mimetype, required)
                    in transform_tool.listPolicies()
                    if mimetype == output_mimetype]
        if policies:
            self.logger.info("Unregistering transform policy "
                             "for type '%s'" % output_mimetype)
            transform_tool.manage_delPolicies([output_mimetype])

    def filter_types_to_registered_ones_only(self, types):
        registry = getToolByName(self.context, 'mimetypes_registry')
        for mimetype in types:
            mts = registry.lookup(mimetype)
            if mts:
                yield mimetype

    def update_input_types(self, transform_name, types):
        transform_tool = getUtility(IPortalTransformsTool)
        transform = transform_tool[transform_name]
        transform.inputs = types[:]
        transform._config['inputs'] = types[:]
        # Update the portal_transform tool's internal _mtmap
        transform_tool._mapTransform(transform)
        self.logger.info(
            "Updated input types for transform '%s'." % transform_name)


def install_tika_policy(util, logger):
    """Install the transform policy for the tika_to_plain_text transform.

    If this policy is already present from an earlier install, remove and
    re-register it. If another policy for this output MIME type is present,
    bail with a decent error message.
    """

    output_mimetype = 'text/plain'
    required_transform = TIKA_TRANSFORM_NAME

    transform_tool = getUtility(IPortalTransformsTool)
    policies = [(mimetype, required) for (mimetype, required)
                in transform_tool.listPolicies()
                if mimetype == output_mimetype]

    if len(policies) == 1 and policies[0][1] == (required_transform, ):
        # That's our own policy (from a previous install)
        # Unregister it before registering it again
        util.unregister_transform_policy(output_mimetype)

    try:
        util.register_transform_policy(output_mimetype, TIKA_TRANSFORM_NAME)
    except TransformException:
        logger.error(
            "There is already a transform policy for '%s' installed! "
            "Please remove it first using the portal_transforms tool "
            "before attempting to install ftw.tika." % output_mimetype)

        raise TransformException(
            "A policy for 'text/plain' already exists - please uninstall "
            "it first if you want to install ftw.tika.")


# Handlers called with DirectoryImportContext by portal_setup tool

def install_portal_transforms(context):
    """Registers portal transforms for Tika integration.
    """
    if context.readDataFile('ftw.tika.default.txt') is None:
        return
    logger = context.getLogger('ftw.tika')
    site = context.getSite()

    util = RegistrationUtility(site, logger)
    util.register_transform(Tika2TextTransform)

    # Filter supported types to only those available in Plone's MIME type
    # registry in order to not cause exceptions during transform registration
    types = list(util.filter_types_to_registered_ones_only(TYPES))
    util.update_input_types(TIKA_TRANSFORM_NAME, types)

    install_tika_policy(util, logger)


def uninstall_portal_transforms(context):
    """Unregisters portal transforms for Tika integration.
    """
    if context.readDataFile('ftw.tika.uninstall.txt') is None:
        return
    logger = context.getLogger('ftw.tika')
    site = context.getSite()

    util = RegistrationUtility(site, logger)
    util.unregister_transform(TIKA_TRANSFORM_NAME)
    util.unregister_transform_policy("text/plain")



from setuptools import setup, find_packages
import os

version = '2.10.0'

tests_require = [
    'Products.CMFCore',
    'ftw.testing',
    'plone.app.testing',
    'plone.testing',
    'testfixtures',
    'unittest2',
    'zope.configuration',
    ]

setup(name='ftw.tika',
      version=version,
      description='Apache Tika integration for Plone using portal transforms.',
      long_description=open('README.rst').read() + '\n' +
      open(os.path.join('docs', 'HISTORY.txt')).read(),

      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'Framework :: Plone :: 5.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Software Development',
        ],

      keywords='plone ftw tika full text indexing apache',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.tika',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'setuptools',
        'requests',

        # Zope
        'zope.component',
        'zope.interface',
        'zope.schema',
        'ZODB3',

        # Plone
        'Plone',
        'Products.GenericSetup',
        'Products.PortalTransforms',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      entry_points='''
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      ''',
      )



from Products.CMFCore.utils import getToolByName
from Products.PortalTransforms.data import datastream
from Products.PortalTransforms.interfaces import ITransform
from ZODB.POSException import ConflictError
from ftw.testing import MockTestCase
from ftw.tika.exceptions import TikaJarNotConfigured
from ftw.tika.testing import FTW_TIKA_INTEGRATION_TESTING
from ftw.tika.tests.utils import RaisingConverter
from ftw.tika.transforms.tika_to_plain_text import Tika2TextTransform
from zope.interface.verify import verifyClass
from zope.interface.verify import verifyObject


class TestTransforms(MockTestCase):

    layer = FTW_TIKA_INTEGRATION_TESTING

    def test_transform_conforms_to_interface(self):
        verifyClass(ITransform, Tika2TextTransform)

        transform = Tika2TextTransform(name='tika_to_plain_text')
        verifyObject(ITransform, transform)

    def test_transform_doesnt_raise_if_jar_path_missing(self):
        portal = self.layer['portal']
        transforms = getToolByName(portal, 'portal_transforms')

        try:
            transforms.convertTo('text/plain',
                                 '',
                                 mimetype='application/pdf')
        except TikaJarNotConfigured, e:
            self.fail("transform raised '%s: %s' unexpectedly!" % (
                    e.__class__.__name__, e))

    def test_transform_doesnt_swallow_conflict_errors(self):
        stream = datastream('dummy')

        # Patch TikaConverter class to just raise a ConflictError
        MockConverter = self.mocker.replace('ftw.tika.converter.TikaConverter')
        self.expect(MockConverter()).result(RaisingConverter(ConflictError))
        self.replay()

        transform = Tika2TextTransform()
        with self.assertRaises(ConflictError):
            transform.convert('', stream)

    def test_transform_doesnt_swallow_keyboard_interrupts(self):
        stream = datastream('dummy')

        # Patch TikaConverter class to just raise a ConflictError
        MockConverter = self.mocker.replace('ftw.tika.converter.TikaConverter')
        self.expect(MockConverter()).result(
            RaisingConverter(KeyboardInterrupt))
        self.replay()

        transform = Tika2TextTransform()
        with self.assertRaises(KeyboardInterrupt):
            transform.convert('', stream)



from ftw.tika.exceptions import TikaJarNotFound
from ftw.tika.testing import FtwTikaLayer
from plone.app.testing import FunctionalTesting
from Products.CMFCore.utils import getToolByName
from unittest2 import TestCase
from zope.configuration import xmlconfig
import os


test_part_dir = os.getcwd()
test_jar_path = os.path.join(test_part_dir, 'tika-app.jar')


class FtwTikaZCMLDirectiveLayer(FtwTikaLayer):

    def setUpZope(self, app, configurationContext):
        super(FtwTikaZCMLDirectiveLayer, self).setUpZope(app,
                                                         configurationContext)
        # Load meta.zcml to register ZCML directive
        import ftw.tika
        xmlconfig.file('meta.zcml', ftw.tika,
                       context=configurationContext)

        # Set JAR path to our temp file using ZCML
        xmlconfig.string(
            '<configure xmlns:tika="http://namespaces.plone.org/tika">'
            '    <tika:config path="%s" />'
            '</configure>' % test_jar_path,
            context=configurationContext)


FTW_TIKA_ZCML_DIRECTIVE_FIXTURE = FtwTikaZCMLDirectiveLayer()
FTW_TIKA_ZCML_DIRECTIVE_FUNCTIONAL = FunctionalTesting(
    bases=(FTW_TIKA_ZCML_DIRECTIVE_FIXTURE,),
    name='FtwTika:FunctionalZCMLDirective')


class TestZCMLDirective(TestCase):

    layer = FTW_TIKA_ZCML_DIRECTIVE_FUNCTIONAL

    def test_transform_uses_path_set_by_zcml_directive(self):
        portal = self.layer['portal']
        transforms = getToolByName(portal, 'portal_transforms')

        # Monkey patch run_process helper
        from ftw.tika import converter
        _run_process = converter.run_process
        converter.run_process = lambda cmd: ('TEXT', 'stderr')

        # Create the file specified by the ZCML directive
        with open(test_jar_path, 'w'):
            stream = transforms.convertTo('text/plain',
                                          '',
                                          mimetype='application/pdf')
            plain_text = stream.getData()
        os.remove(test_jar_path)

        self.assertEquals(plain_text, 'TEXT')

        # Restore the original run_process function
        converter.run_process = _run_process

    def test_invalid_jar_path_in_zcml_doesnt_cause_transform_to_raise(self):
        portal = self.layer['portal']
        transforms = getToolByName(portal, 'portal_transforms')

        # DON'T create the file specified by the ZCML directive
        try:
            transforms.convertTo('text/plain',
                                 '',
                                 mimetype='application/pdf')
        except TikaJarNotFound, e:
            self.fail("transform raised '%s: %s' unexpectedly!" % (
                    e.__class__.__name__, e))



from StringIO import StringIO
from ftw.testing import MockTestCase
from ftw.tika.converter import TikaConverter
from ftw.tika.exceptions import ProcessError
from ftw.tika.exceptions import TikaConversionError
from ftw.tika.exceptions import TikaJarNotConfigured
from ftw.tika.exceptions import TikaJarNotFound
from ftw.tika.interfaces import IZCMLTikaConfig
from ftw.tika.testing import FTW_TIKA_INTEGRATION_TESTING
from mocker import ARGS
from zope.component import getGlobalSiteManager
from zope.component import getUtility
import tempfile


class TestConverter(MockTestCase):

    layer = FTW_TIKA_INTEGRATION_TESTING

    def test_converter_builds_correct_command_line(self):
        # Patch run_process to just return stderr and the command line given

        def return_cmd_line(cmd):
            return (cmd, '')

        mock_run_proc = self.mocker.replace('ftw.tika.converter.run_process')
        self.expect(mock_run_proc(ARGS)).call(return_cmd_line)
        self.replay()

        jar_path = '/bin/ls'
        tika_converter = TikaConverter(path=jar_path)
        cmd = tika_converter.convert('')
        cmd_without_doc_filename = cmd.split()[:-1]

        self.assertEquals(cmd_without_doc_filename,
                          ['java', '-jar', jar_path, '-t'])

    def test_converter_accepts_file_like_stream_object(self):
        sample_text = 'TEXT'

        # Patch run_process to just return sample output
        mock_run_proc = self.mocker.replace('ftw.tika.converter.run_process')
        self.expect(mock_run_proc(ARGS)).result((sample_text, 'stderr'))
        self.replay()

        with tempfile.NamedTemporaryFile() as tmp_file:
            tika_converter = TikaConverter(path=tmp_file.name)
            plain_text = tika_converter.convert(StringIO(sample_text))

        self.assertEquals(plain_text, sample_text)

    def test_process_error_causes_coverter_to_raise_conversion_error(self):
        # Patch run_process to just raise a ProcessError

        def raise_process_error(cmd):
            raise ProcessError

        mock_run_proc = self.mocker.replace('ftw.tika.converter.run_process')
        self.expect(mock_run_proc(ARGS)).call(raise_process_error)
        self.replay()

        with tempfile.NamedTemporaryFile() as tmp_file:
            tika_converter = TikaConverter(path=tmp_file.name)
            with self.assertRaises(TikaConversionError):
                tika_converter.convert('')

    def test_missing_jar_path_causes_converter_to_raise(self):
        # Since the path is configured in the IZCMLTikaConfig from the ZCML
        # loaded in the testing layer, we need to unregister the config for
        # this test to verify the exception.
        config = getUtility(IZCMLTikaConfig)
        getGlobalSiteManager().unregisterUtility(provided=IZCMLTikaConfig)
        try:
            with self.assertRaises(TikaJarNotConfigured):
                TikaConverter().convert('')
        finally:
            getGlobalSiteManager().registerUtility(component=config)

    def test_invalid_jar_path_causes_converter_to_raise(self):
        tika_converter = TikaConverter(path="/nonexistent")
        with self.assertRaises(TikaJarNotFound):
            tika_converter.convert('')
